import functools
import logging
import uuid

import gromozeka
import gromozeka.concurrency
import gromozeka.concurrency.scheduler
from gromozeka.primitives import TaskType, SchedulerType, Request

PENDING = 'PENDING'
RECEIVED = 'RECEIVED'
STARTED = 'STARTED'
SUCCESS = 'SUCCESS'
FAILURE = 'FAILURE'
CANCELED = 'CANCELED'
REJECTED = 'REJECTED'
RETRY = 'RETRY'


def task(bind=False, max_retries=0, retry_countdown=1, eta=None, every=None, interval=None, at=None):
    """Task decorator. Use to make usual function as task

    After this, you must register task

    Examples:

        Register task:

        .. code-block:: python

            @task()
            def hello_world():
                print('hello world')

            app=ATP()

            app.register_task(
                hello_world(),
                broker_point=BrokerPointType(
                    exchange="telegram",
                    exchange_type='direct',
                    queue='fib',
                    routing_key='fib'))

            app.start()

        or :obj:`Task` object can register self:

        .. code-block:: python

            @task()
            def hello_world():
                print('hello world')

            app=ATP()

            hello_world().register(
                broker_point=BrokerPointType(
                    exchange="hello",
                    exchange_type='direct',
                    queue='hello',
                    routing_key='hello'))

            app.start()

    Args:
        bind(:obj:`bool`, optional): If `True` `Task` will be in function arguments as first argument
        max_retries(:obj:`int`, optional): Maximum number of retries
        retry_countdown(:obj:`int` or :obj:`float`, optional): Pause between retries
        eta(str): Datetime when to run task, formatted as "%Y-%m-%d %H:%M:%S.%f"
        every(:obj:`str`, optional): May be seconds, minutes, hours, days, weeks.
        interval(:obj:`int`, optional): Uses with `every`.
        at(:obj:`str`, optional): Time formatted as `hour:minute` or `hour:minute:second`.

    Returns:
        gromozeka.primitives.Task: task object
    """

    def task_decor(f):
        @functools.wraps(f)
        def wrapper():
            return Task(func=f, bind=bind, max_retries=max_retries, retry_countdown=retry_countdown, eta=eta,
                        every=every, interval=interval, at=at)

        return wrapper

    return task_decor


class RegistryTask:
    """Task registry class

    Attributes:
        id(:obj:`str`): Task identification
        func: Task function
        bind(:obj:`bool`): If `True` `Task` will be in function arguments as first argument
        broker_point(:obj:`gromozeka.primitives.BrokerPointType`): Broker entry
        worker_class(:obj:`gromozeka.concurrency.Worker`): Task workers will be this type
        max_workers(:obj:`int`): Number of task workers
        max_retries(:obj:`int`): Maximum number of retries
        retry_countdown(:obj:`int` or :obj:`float`): Pause between retries
        pool(:obj:`gromozeka.concurrency.Pool`): Task workers pool

    Args:
        broker_point(gromozeka.primitives.BrokerPointType): Broker entry
        id_(str): Task identification
        func: Task function
        bind(bool): If `True` `Task` will be in function arguments as first argument
        worker_class(gromozeka.concurrency.Worker): Task workers will be this type
        max_workers(int): Number of task workers
        retry_countdown(:obj:`int` or :obj:`float`): Pause between retries
    """

    __slots__ = ['id', 'func', 'bind', 'broker_point', 'worker_class', 'max_workers', 'max_retries', 'retry_countdown',
                 'pool']

    def __init__(self, broker_point, id_, func, bind, worker_class, max_workers, max_retries, retry_countdown):
        self.id = id_
        self.func = func
        self.bind = bind
        self.broker_point = broker_point
        self.worker_class = worker_class or gromozeka.concurrency.ThreadWorker
        self.max_workers = max_workers
        self.max_retries = max_retries
        self.retry_countdown = retry_countdown
        self.pool = gromozeka.concurrency.Pool(max_workers=self.max_workers, worker_class=self.worker_class)


class Task:
    """Task class.

    Attributes:
        uuid(:obj:`uuid.uuid4`, optional): Unique task runtime identification
        id(:obj:`str`): Task registry identification
        logger(:obj:`logging.Logger`): Class logger
        state(:obj:`str`): Current task state
        func: Task function
        args(:obj:`list`):Task function arguments
        kwargs(:obj:`dict`): Task function keyword arguments
        bind(:obj:`bool`): If `True` `Task` will be in function arguments as first argument
        max_retries(:obj:`int`): Maximum number of retries
        retry_countdown(:obj:`int` or :obj:`float`): Pause between retries
        retries(:obj:`int`): retries counter
        countdown(:obj:`str`): Retry countdown (%Y-%m-%d %H:%M:%S.%f)
        eta(:obj:`str`, optional): Task ETA (%Y-%m-%d %H:%M:%S.%f)
        broker_point(:obj:`gromozeka.primitives.BrokerPointType`): Broker entry
        delivery_tag(:obj:`int`): Original message delivery tag

    Args:
        func: Task function
        args(:obj:`list`, optional):Task function arguments
        kwargs(:obj:`dict`, optional): Task function keyword arguments
        bind(bool): If `True` `Task` will be in function arguments as first argument
        state(:obj:`str`, optional): Current task state
        max_retries(:obj:`int`, optional): Maximum number of retries
        retry_countdown(:obj:`int` or :obj:`float`, optional): Pause between retries
        retries(:obj:`int`, optional): retries counter
        countdown(:obj:`str`, optional): Retry countdown (%Y-%m-%d %H:%M:%S.%f)
        eta(:obj:`str`, optional): Task ETA (%Y-%m-%d %H:%M:%S.%f)
        uuid_(:obj:`uuid.uuid4`, optional): Unique task runtime identification
        broker_point(gromozeka.primitives.BrokerPointType): Broker entry
        delivery_tag(:obj:`int`, optional): Original message delivery tag
    """

    __slots__ = ['uuid', 'id', 'logger', 'func', 'args', 'kwargs', 'bind', 'max_retries', 'retry_countdown', 'retries',
                 'countdown', '_eta', 'broker_point', 'delivery_tag', 'state', '_every', '_interval', '_at']

    def __init__(self, func, args=None, kwargs=None, bind=None, state=PENDING, max_retries=0, retry_countdown=1,
                 retries=0, countdown='', eta='', uuid_=None, broker_point=None, delivery_tag=None, every=None,
                 interval=None, at=None):

        self.uuid = uuid_ or uuid.uuid4()
        self.id = "%s.%s" % (func.__module__, func.__name__)
        self.logger = logging.getLogger("gromozeka.pool.worker.{}".format(self.id))
        self.state = state
        self.func = func
        self.args = args or []
        self.kwargs = kwargs or {}
        self.bind = bind or False
        self.max_retries = max_retries
        self.retry_countdown = retry_countdown
        self.retries = retries
        self.countdown = countdown
        self._eta = eta
        self.broker_point = broker_point
        self.delivery_tag = delivery_tag
        self._every = every
        self._interval = interval
        self._at = at

    def register(self, broker_point, worker_class=None, max_workers=1, max_retries=0, retry_countdown=0):
        """Register task in ATP task registry

        Args:
            broker_point(gromozeka.primitives.BrokerPointType): Broker entry
            worker_class(gromozeka.concurrency.Worker): Task workers will be this type
            max_workers(int): Number of task workers
            max_retries(int): Maximum number of retries
            retry_countdown(:obj:`int` or :obj:`float`): Pause between retries

        Returns:
            gromozeka.primitives.Task:
        """
        self.broker_point = broker_point
        r_task = RegistryTask(id_=self.id, func=self.func, bind=self.bind, worker_class=worker_class,
                              max_workers=max_workers, max_retries=max_retries or self.max_retries,
                              retry_countdown=retry_countdown or self.retry_countdown or 1,
                              broker_point=broker_point or self.broker_point)
        atp = gromozeka.get_app()
        atp.register_task(task=r_task, broker_point=self.broker_point)
        atp.broker.add_consumer(task_id=self.id, broker_point=broker_point)
        return self

    def __call__(self, *args, **kwargs):
        """Execute `Task` function

        Examples:

            .. code-block:: python

                @task()
                def some_task(a,b):
                    print(a+b)

                some_task()(1,2)

        Args:
            *args: function arguments
            **kwargs: function keyword arguments

        Returns:
            function result
        """
        return functools.partial(self.func, *(self, *args), **kwargs)()

    @property
    def request(self):
        """Task `Request` property

        Returns:
            gromozeka.primitives.Request:
        """
        task_ = TaskType(uuid=self.uuid.__str__(), id=self.id, args=self.args, kwargs=self.kwargs, retries=self.retries,
                         state=self.state, delivery_tag=self.delivery_tag)
        if self._eta or self.countdown:
            scheduler = SchedulerType(countdown=self.countdown, eta=self._eta, every=self._every,
                                      interval=self._interval, at=self._at)
        else:
            scheduler = None
        return Request(task=task_, broker_point=self.broker_point, scheduler=scheduler)

    @classmethod
    def from_request(cls, request):
        """Make `Task` from `Request`

        Args:
            request(gromozeka.primitives.Request): task `Request`

        Returns:
            gromozeka.primitives.Task:
        """
        atp = gromozeka.get_app()
        r_task = atp.get_task(request.task.id)
        return cls(func=r_task.func,
                   args=request.task.args,
                   kwargs=request.task.kwargs,
                   bind=r_task.bind,
                   max_retries=r_task.max_retries,
                   retry_countdown=r_task.retry_countdown,
                   retries=request.task.retries,
                   countdown=request.scheduler.countdown if request.scheduler else None,
                   eta=request.scheduler.eta if request.scheduler else None,
                   every=request.scheduler.every if request.scheduler else None,
                   interval=request.scheduler.interval if request.scheduler else None,
                   at=request.scheduler.at if request.scheduler else None,
                   uuid_=request.task.uuid,
                   broker_point=request.broker_point,
                   delivery_tag=request.task.delivery_tag)

    def apply_async(self, *args, **kwargs):
        """Run `Task` asynchronously

        Examples:

            .. code-block:: python

                @task()
                def some_task(a,b):
                    print(a+b)

                some_task().apply_async(1,2)

        Args:
            *args: `Task` function arguments
            **kwargs: `Task` function keyword arguments
        """
        if self._every:
            delay = gromozeka.concurrency.scheduler.delay(last_run=self._eta, every=self._every,
                                                          interval=self._interval, at=self._at)
            self._eta = gromozeka.concurrency.scheduler.new_eta(delay)
        atp = gromozeka.get_app()
        r_task = atp.get_task(self.id)
        self.broker_point = r_task.broker_point
        self.args = args or self.args
        self.kwargs = kwargs or self.kwargs
        atp.broker.send_task(request=self.request)

    def eta(self, datetime):
        """

        Args:
            datetime(str): Datetime formatted as "%Y-%m-%d %H:%M:%S.%f"

        Returns:
            gromozeka.primitives.Task:
        """
        self._eta = datetime
        return self

    def every(self, period):
        """

        Args:
            period(str): May be seconds, minutes, hours, days, weeks.

        Returns:
            gromozeka.primitives.Task:
        """
        self._every = period
        return self

    def interval(self, count):
        """

        Args:
            count(int): Period interval. Uses with every.

        Returns:
            gromozeka.primitives.Task:
        """
        self._interval = count
        return self

    def at(self, time):
        """

        Args:
            time(str): Time formatted as `hour:minute` or `hour:minute:second`.

        Returns:
            gromozeka.primitives.Task:
        """
        self._at = time
        return self

    def on_receive(self):
        """Event when task received in worker

        """
        self.state = RECEIVED

    def on_start(self):
        """Event when started tu run task

        """
        self.state = STARTED

    def on_reject(self):
        """Event when task rejected

        """
        self.state = REJECTED
        self._reject()

    def on_cancel(self):
        """Event when task canceled

        """
        self.state = CANCELED
        self._reject()

    def on_retry(self, e):
        """Event when task on retry

        Args:
            e(gromozeka.exceptions.Retry):

        Returns:
             bool: `True` if task will retry more times
        """

        if not e.max_retries:
            e.max_retries = self.max_retries
        if not e.retry_countdown:
            e.retry_countdown = self.retry_countdown
        if e.max_retries <= self.retries:
            return False
        self.state = RETRY
        self._retry(e)
        return True

    def on_success(self):
        """Event when task finishes successfully

        """
        if self.state == CANCELED:
            return
        self.state = SUCCESS
        self.retries = 0
        self._ack()
        if self._every:
            self.uuid = uuid.uuid4()
            self.apply_async(*self.args, **self.kwargs)

    def on_fail(self, e):
        """Event when task failed with `Exception`

        """
        self.state = FAILURE
        self._reject()
        raise e

    def cancel(self):
        """Cancel task. Use this method in `Task` function. Bind option must be `True`

        Examples:
            .. code-block:: python

                @task(bind=True)
                def some_task():
                    try:
                        a = 1 / 0
                    except ZeroDivisionError as e:
                        self.cancel()
        """
        self.on_cancel()

    def _retry(self, e):
        """

        Args:
            e(gromozeka.exceptions.Retry):

        Returns:

        """
        self.logger.warning("retry on error: {0}".format(e))
        self.countdown = gromozeka.concurrency.scheduler.new_eta(seconds=e.retry_countdown)
        self.retries += 1
        atp = gromozeka.get_app()
        atp.broker.send_task(request=self.request)
        self._reject()

    def _ack(self):
        atp = gromozeka.get_app()
        atp.broker.ack(broker_point=self.broker_point, delivery_tag=self.delivery_tag)

    def _reject(self):
        atp = gromozeka.get_app()
        atp.broker.reject(broker_point=self.broker_point, delivery_tag=self.delivery_tag)



"""Chained tasks

Notes:
    not ready yet

"""
import uuid


class Chain(list):
    def __init__(self):
        self.uuid = uuid.uuid4()
        super().__init__()

    def __or__(self, other):
        """

        Args:
            other(Task):

        Returns:

        """
        [task.chain.append(other) for task in self]
        other.parent = self[-1].uuid
        self.append(other)
        return self

    def __call__(self, *args, **kwargs):
        pass



import logging
import signal
import sys

from gromozeka.brokers import Broker
from gromozeka.concurrency import Scheduler
from gromozeka.config import app_config
from gromozeka.exceptions import GromozekaException
from gromozeka.primitives import RegistryTask

DEFAULT_LOGGER_FORMAT = '%(asctime)-15s %(levelname)s %(name)s %(threadName)s %(processName)s %(message)s'


def get_app():
    """

    Returns:
         gromozeka.app.Gromozeka:
    """
    try:
        return atp
    except NameError:
        logging.getLogger("ATP").error('async task processor (ATP) not initialized')
        sys.exit(1)


class Gromozeka:
    """

    Attributes:
        logger(:obj:`logging.Logger`): Class logger
        broker(:obj:`gromozeka.brokers.Broker`): Broker
        scheduler(:obj:`gromozeka.concurrency.Scheduler`): Scheduler
        registry(:obj:`dict` of :obj:`gromozeka.primitives.Task`): Task registry
        config(:obj:`gromozeka.config.Config`): config object
    """

    __slots__ = ['logger', 'broker', 'scheduler', 'registry', 'config']

    def __init__(self):
        global atp
        atp = self
        self.config = app_config
        self.logger = logging.getLogger('gromozeka')
        self.broker = Broker(self)
        self.scheduler = Scheduler()
        self.registry = {}

    def config_from_env(self):
        """Configure Gromozeka with environment variables

        Examples:
            .. code-block:: python

                app=Gromozeka().config_from_env()

        Returns:
            gromozeka.Gromozeka: Configured application
        """
        self.config.from_env()
        return self

    def config_from_dict(self, conf):
        """Configure Gromozeka with :obj:`dict`

        Examples:
            .. code-block:: python

                conf={'app_prefix':'my_application','broker_reconnect_max_retries':3}
                app=Gromozeka().config_from_dict(conf)

        Args:
            conf(dict): config :obj:`dict`

        Returns:
            gromozeka.Gromozeka: Configured application
        """
        self.config.from_dict(conf)
        return self

    def start(self):
        """Start application

        """
        # setup signal handler
        for sig in ('SIGTERM', 'SIGINT'):
            signal.signal(getattr(signal, sig), self._signal_handler)

        self.broker.start()
        self.scheduler.start()
        for name, task in self.registry.items():
            task.pool.start()

    def get_consumer(self, id_):
        """Get task consumer by task_id

        Args:
            id_(str): Unique task identification

        Returns:
             gromozeka.brokers.Consumer:
        """
        return self.broker.get_consumer(id_)

    def register_task(self, task, broker_point, worker_class=None, max_workers=1, max_retries=0, retry_countdown=0):
        """

        Args:
            task(:obj:`gromozeka.primitives.Task` or :obj:`gromozeka.primitives.RegistryTask`): Task
            broker_point(gromozeka.primitives.BrokerPointType): Broker entry
            worker_class(:class:`gromozeka.concurrency.Worker`, optional): Worker class
            max_workers(int): How match workers will start
            max_retries(int): Maximum number of retries, after it will reached task will down
            retry_countdown(int): Pause between retries (seconds)
        """
        if isinstance(task, RegistryTask):
            self.registry[task.id] = task
        else:
            task.register(broker_point=broker_point, worker_class=worker_class, max_workers=max_workers,
                          max_retries=max_retries, retry_countdown=retry_countdown)

    def get_task(self, id_):
        """Get task by id

        Args:
            id_(str): Unique task identification

        Returns:
            gromozeka.primitives.RegistryTask: registry task
        """
        try:
            return self.registry[id_]
        except KeyError:
            raise GromozekaException('task `{}` not registered in ATP registry'.format(id_))

    def _signal_handler(self, _, __):
        """Signal handler

        """
        self.stop()

    def stop(self):
        self.logger.debug('trying to shutdown graceful')
        self.scheduler.stop()
        for uuid, task in self.registry.items():
            task.pool.stop()
        self.broker.stop()



import logging
import queue
import time

import pika
from pika.exceptions import ChannelClosed

from gromozeka.concurrency import Pool
from gromozeka.concurrency import ThreadWorker
from gromozeka.concurrency import commands
from gromozeka.concurrency import scheduler
from gromozeka.config import app_config
from gromozeka.primitives.protocol import BrokerPointType, Request

CONSUMER_ID_FORMAT = "{exchange}_{exchange_type}_{queue}_{routing_key}"
SCHEDULER_CONSUMER_NAME = 'scheduler'
ETA_CONSUMER_NAME = 'eta'


# TODO extract to universal (BaseConsumer) - this is replacement solution
class Consumer:
    """Channel consumer

    Attributes:
        logger(:obj:`logging.Logger`): Class logger
        broker(:obj:`gromozeka.brokers.Consumer`): Broker
        exchange(:obj:`str`): Rabbitmq exchange
        exchange_type(:obj:`str`): : Rabbitmq exchange type
        queue_(:obj:`str`): Rabbitmq queue
        routing_key(:obj:`str`): : Rabbitmq routing key
        id(:obj:`str`): Consumer identification
        tasks(:obj:`dict` of :obj:`gromozeka.primitives.RegistryTask`): Consumer tasks
        prefetch_count(:obj:`int`): prefetch_count size
        _connection(:obj:`pika.SelectConnection`): pika connection
        _channel: Channel
        _closing(:obj:`bool`): Consumer state
        _consumer_tag(:obj:`int`): Consumer tag

    Args:
        broker(gromozeka.brokers.Broker): Broker
        connection(pika.SelectConnection): pika connection
        exchange(str): Rabbitmq exchange
        exchange_type(str): : Rabbitmq exchange type
        queue_(str): : Rabbitmq queue
        routing_key(str): : Rabbitmq routing key
    """

    def __init__(self, broker, connection, exchange, exchange_type, queue_, routing_key):
        self.broker = broker
        self.logger = logging.getLogger('gromozeka.broker.consumer')
        self.prefetch_count = 0
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.queue = queue_
        self.routing_key = routing_key
        self.id = CONSUMER_ID_FORMAT.format(
            exchange=self.exchange,
            exchange_type=self.exchange_type,
            queue=self.queue,
            routing_key=self.routing_key)
        self._channel = None
        self._connection = connection
        self._closing = False
        self._consumer_tag = None
        self.tasks = {}

    @staticmethod
    def format_consumer_id_from_broker_point(broker_point):
        """Format consumer id from `primitives.BrokerPointType`

        Args:
            broker_point(gromozeka.BrokerPointType): Broker entry

        Returns:
            str: consumer identification
        """
        return CONSUMER_ID_FORMAT.format(
            exchange=broker_point.exchange,
            exchange_type=broker_point.exchange_type,
            queue=broker_point.queue,
            routing_key=broker_point.routing_key)

    def get_broker_point(self):
        """Get broker point

        Returns:
            gromozeka.BrokerPointType: Broker entry
        """
        return BrokerPointType(
            exchange=self.exchange,
            exchange_type=self.exchange_type,
            queue=self.queue,
            routing_key=self.routing_key)

    def add_task(self, task_id):
        """Add task to consumer queue

        Args:
            task_id(str): Registry task identification
        """
        self.tasks[task_id] = self.broker.app.get_task(task_id)

    def open_channel(self):
        """Open channel

        """
        self.logger.debug('Creating a new channel')
        self._connection.channel(on_open_callback=self._on_channel_open)

    def on_message(self, _, basic_deliver, __, body):
        """Event when message received

        Args:
            _: channel
            basic_deliver:
            __:
            body:
        """
        self.logger.info('Received message # %s from %s: %s', basic_deliver.delivery_tag, self.id, body)
        request = Request.from_json(json_message=body, delivery_tag=basic_deliver.delivery_tag)
        if self.queue in (self.broker.config.broker_retry_queue, self.broker.config.broker_eta_queue):
            if request.scheduler.every and not request.scheduler.eta:  # raw task
                delay = scheduler.delay(last_run=request.scheduler.eta,
                                        every=request.scheduler.every,
                                        interval=request.scheduler.interval,
                                        at=request.scheduler.at)
            else:
                delay = scheduler.delay_from_eta(
                    eta=request.scheduler.countdown if self.queue == self.broker.config.broker_retry_queue else request.scheduler.eta)
            if delay > 0:
                self.broker.app.scheduler.add(
                    wait_time=delay,
                    func=self.publish,
                    args=[request, request.broker_point.exchange, request.broker_point.routing_key],
                    priority=1 if self.queue == self.broker.config.broker_retry_queue else 2)
            else:
                self.publish(request, request.broker_point.exchange, request.broker_point.routing_key)
            self.acknowledge_message(basic_deliver.delivery_tag)
        else:
            request = Request.from_json(json_message=body, delivery_tag=basic_deliver.delivery_tag)
            if not request:
                return self.acknowledge_message(basic_deliver.delivery_tag)
            self.tasks[request.task.id].pool.in_queue.put(request)

    def publish(self, request, exchange=None, routing_key=None):
        """Publish request message to channel channel

        Args:
            request (gromozeka.primitives.Request):
            exchange(:obj:`str`, optional):
            routing_key(:obj:`str`, optional):
        """
        self._channel.basic_publish(exchange or self.exchange, routing_key or self.routing_key, request.to_json())

    def acknowledge_message(self, delivery_tag):
        """Acknowledge message by it`s delivery tag

        Args:
            delivery_tag(int):
        """
        self.logger.info('Acknowledging message %s', delivery_tag)
        self._channel.basic_ack(delivery_tag)

    def reject_message(self, delivery_tag, requeue=False):
        """Reject message by it`s delivery tag

        Args:
            delivery_tag(int):
            requeue(bool): If `True` message will reject with requeue
        """
        self.logger.info('Rejecting message %s', delivery_tag)
        self._channel.basic_reject(delivery_tag, requeue)

    def change_prefetch_count(self, new_size):
        if self._channel:
            self._channel.basic_qos(prefetch_count=new_size)
            self.prefetch_count = new_size
        else:
            self.prefetch_count = new_size

    def stop_consuming(self):
        """Stop consuming

        """
        if self._channel:
            self.logger.debug('Sending a Basic.Cancel RPC command to RabbitMQ')
            self._channel.basic_cancel(self._on_cancelok, self._consumer_tag)

    def add_on_cancel_callback(self):
        self.logger.debug('Adding consumer cancellation callback')
        self._channel.add_on_cancel_callback(self._on_consumer_cancelled)

    def _on_channel_open(self, channel):
        self.logger.debug('Channel opened')
        self._channel = channel
        self._add_on_channel_close_callback()
        self._setup_exchange(self.exchange)

    def _add_on_channel_close_callback(self):
        self.logger.debug('Adding channel close callback')
        self._channel.add_on_close_callback(self._on_channel_closed)

    def _on_channel_closed(self, channel, reply_code, reply_text):
        self.logger.debug('Channel %i was closed: (%s) %s',
                          channel, reply_code, reply_text)
        # self._connection.close()

    def _setup_exchange(self, exchange_name):
        self.logger.debug('Declaring exchange %s', exchange_name)
        self._channel.exchange_declare(self._on_exchange_declareok,
                                       exchange_name,
                                       self.exchange_type)

    def _on_exchange_declareok(self, _):
        self.logger.debug('Exchange declared')
        self._setup_queue(self.queue)

    def _setup_queue(self, queue_name):
        self.logger.debug('Declaring queue %s', queue_name)
        self._channel.queue_declare(self._on_queue_declareok, queue_name)

    def _on_queue_declareok(self, _):
        self.logger.debug('Binding %s to %s with %s',
                          self.exchange, self.queue, self.routing_key)
        self._channel.queue_bind(self._on_bindok, self.queue,
                                 self.exchange, self.routing_key)

    def _on_bindok(self, _):
        self.logger.debug('Queue bound')
        self._start_consuming()

    def _start_consuming(self):
        self.logger.debug('Issuing consumer related RPC commands')
        self._channel.basic_qos(prefetch_count=self.prefetch_count)
        self.add_on_cancel_callback()
        self._consumer_tag = self._channel.basic_consume(self.on_message, self.queue)

    def _on_consumer_cancelled(self, method_frame):
        self.logger.debug('Consumer was cancelled remotely, shutting down: %r',
                          method_frame)
        if self._channel:
            self._channel.close()

    def _on_cancelok(self, _):
        self.logger.debug('RabbitMQ acknowledged the cancellation of the consumer')
        self._close_channel()

    def _close_channel(self):
        self.logger.debug('Closing the channel')
        self._channel.close()


# TODO extract to universal (BaseBroker) - this is replacement solution
class Broker(Pool):
    """Broker class

    Attributes:
        app(:obj:`gromozeka.Gromozeka`): Application
        logger(:obj:`logging.Logger`): Class logger
        config(gromozeka.config.Config): Config object
        `Worker`(:obj:`gromozeka.concurrency.Worker`): Parent class for `Customer`
        _url(:obj:`str`): Rabbitmq url
        _connection(:obj:`pika.SelectConnection`): pika connection
        _closing(:obj:`bool`): Broker state
        _consumers(:obj:`list` of :obj:`Consumer`): consumers

    """

    class Worker(ThreadWorker):
        retries = 0
        broker = None

    def __init__(self, app):
        self.app = app
        self.logger = logging.getLogger("gromozeka.broker")
        self.config = None
        self._url = None
        self.retries = 0
        worker = self.Worker
        worker.run = self.worker_run
        worker.broker = self
        self._connection = None
        self._closing = False
        self._consumers = {}
        super().__init__(max_workers=1, worker_class=worker, logger=self.logger)

    def listen_cmd(self):
        """Listen to commands

        """
        while not self._stop_event.is_set():
            try:
                command, args = self.cmd_in_queue.get(True, 0.05)
            except queue.Empty:
                continue
            if command == commands.POOL_GROW:
                self._grow(args)
            elif command == commands.POOL_STOP:
                self.stop_()
            elif command == commands.BROKER_ADD_CONSUMER:
                self._add_consumer(task_id=args['task_id'], broker_point=args['broker_point'])
            elif command == commands.BROKER_PUBLISH:
                self._send_task(args)
            elif command == commands.BROKER_ACK:
                self._ack(broker_point=args['broker_point'], delivery_tag=args['delivery_tag'])
            elif command == commands.BROKER_REJECT:
                self._reject(broker_point=args['broker_point'], delivery_tag=args['delivery_tag'])
            elif command == commands.BROKER_PREFETCH_COUNT:
                self._prefetch_count(args['broker_point'], new_size=args['new_size'])

    @staticmethod
    def worker_run(self):
        """Overrides `Worker` method

        Args:
            self(gromozeka.brokers.rabbit.Broker.Worker): Parent `Worker` object
        """
        self.logger.info("start")
        while not self._stop_event.is_set():
            try:
                self.broker.serve()
                self.retries = 0
            except pika.exceptions.AMQPConnectionError:
                pass
            if self._stop_event.is_set():
                break
            if self.retries == self.broker.config.broker_reconnect_max_retries:
                self.broker.app.stop()
                break
            self.retries += 1
            time.sleep(self.broker.config.broker_reconnect_retry_countdown)
            self.logger.info('Trying to reconnect')
        self.pool.out_queue.cancel_join_thread()
        self.logger.info("stop")

    def run(self):
        """Run `Broker` in thread or process

        """
        self.logger.info("start")
        self.grow(self.init_max_workers)
        self.listen_cmd()

    def serve(self):
        """Start `Broker`

        """
        self.config = app_config
        self._url = self.config.broker_url

        # add scheduler consumer
        self._add_consumer(
            broker_point=BrokerPointType(
                exchange=self.config.broker_retry_exchange,
                exchange_type='direct',
                queue=self.config.broker_retry_queue,
                routing_key=self.config.broker_retry_routing_key),
            name=SCHEDULER_CONSUMER_NAME)

        # add consumer for eta tasks
        self._add_consumer(
            broker_point=BrokerPointType(
                exchange=self.config.broker_eta_exchange,
                exchange_type='direct',
                queue=self.config.broker_eta_queue,
                routing_key=self.config.broker_eta_routing_key),
            name=ETA_CONSUMER_NAME)

        self._connection = self.connect()
        self._connection.ioloop.start()

    def send_task(self, request):
        """Send request to consumer

        Args:
            request(gromozeka.primitives.Request): `Request` object
        """
        self.cmd.put(commands.broker_publish(request=request))

    def _send_task(self, request):
        """

        Args:
            request(gromozeka.primitives.Request): `Request` object
        """
        if request.scheduler:
            if request.scheduler.countdown:
                consumer_id = SCHEDULER_CONSUMER_NAME
            else:
                consumer_id = ETA_CONSUMER_NAME
        else:
            consumer_id = Consumer.format_consumer_id_from_broker_point(request.broker_point)

        self._consumers[consumer_id].publish(request=request)

    def ack(self, broker_point, delivery_tag):
        """Acknowledge message with specific consumer by it's tag

        Args:
            broker_point(gromozeka.primitives.BrokerPointType): Broker entry
            delivery_tag(int): Original message delivery tag
        """
        self.cmd.put(commands.broker_ack(broker_point=broker_point, delivery_tag=delivery_tag))

    def _ack(self, broker_point, delivery_tag):
        consumer_id = Consumer.format_consumer_id_from_broker_point(broker_point=broker_point)
        consumer = self.get_consumer(consumer_id=consumer_id)
        consumer.acknowledge_message(delivery_tag=delivery_tag)

    def reject(self, broker_point, delivery_tag):
        """Reject message with specific consumer by it's tag

        Args:
            broker_point(gromozeka.primitives.BrokerPointType): Broker entry
            delivery_tag(int): Original message delivery tag
        """
        self.cmd.put(commands.broker_reject(broker_point=broker_point, delivery_tag=delivery_tag))

    def _reject(self, broker_point, delivery_tag):
        consumer_id = Consumer.format_consumer_id_from_broker_point(broker_point=broker_point)
        consumer = self.get_consumer(consumer_id=consumer_id)
        consumer.reject_message(delivery_tag=delivery_tag)

    def on_pool_size_change(self):
        for cname, consumer in self._consumers.items():
            new_prefetch_count = 0
            for tname, task in self.app.registry.items():
                if Consumer.format_consumer_id_from_broker_point(task.broker_point) == cname:
                    new_prefetch_count += len(task.pool.workers)
            if consumer.prefetch_count != new_prefetch_count:
                self.logger.info('prefetch count changed from %s to %s' % (consumer.prefetch_count, new_prefetch_count))
                try:
                    consumer.change_prefetch_count(new_prefetch_count)
                except ChannelClosed:
                    consumer.open_channel()
                    consumer.change_prefetch_count(new_prefetch_count)
                if consumer.prefetch_count == 0:
                    consumer.stop_consuming()

    def prefetch_count(self, broker_point, new_size):
        self.cmd.put(commands.broker_prefetch_count(broker_point=broker_point, new_size=new_size))

    def _prefetch_count(self, broker_point, new_size):
        consumer = self.get_consumer(consumer_id=Consumer.format_consumer_id_from_broker_point(broker_point))
        consumer.change_prefetch_count(new_size)

    def get_consumer(self, consumer_id):
        """Get consumer by it's id

        Args:
            consumer_id(str): consumer identification

        Returns:
            gromozeka.brokers.Consumer:
        """
        return self._consumers[consumer_id]

    def stop_(self):
        self._closing = True
        self._connection.close()
        super().stop_()

    def add_consumer(self, broker_point, task_id=None):
        """Add new `Consumer` to `Broker`

        Args:
            broker_point(gromozeka.primitives.BrokerPointType): broker entry
            task_id(str): task identification
        """
        self.cmd.put(commands.broker_add_consumer(task_id=task_id, broker_point=broker_point))

    def _add_consumer(self, broker_point, task_id=None, name=None):
        """

        Args:
            broker_point(gromozeka.primitives.BrokerPointType): Broker entry
            task_id(str): Task identification
            name(str): Consumer name
        """
        if name:
            consumer_id = name
        else:
            consumer_id = CONSUMER_ID_FORMAT.format(
                exchange=broker_point.exchange,
                exchange_type=broker_point.exchange_type,
                queue=broker_point.queue,
                routing_key=broker_point.routing_key)
        try:
            consumer = self._consumers[consumer_id]
        except KeyError:
            consumer = Consumer(
                broker=self,
                connection=self._connection,
                exchange=broker_point.exchange,
                exchange_type=broker_point.exchange_type,
                queue_=broker_point.queue,
                routing_key=broker_point.routing_key)

        if task_id:
            consumer.add_task(task_id)
        self._consumers[consumer_id] = consumer

    def connect(self):
        """Connect to Rabbitmq

        Returns:
            pika.SelectConnection:
        """
        self.logger.debug('Connecting to %s', self._url)
        return pika.SelectConnection(pika.URLParameters(self._url), self._on_connection_open, stop_ioloop_on_close=True)

    def _on_connection_open(self, _):
        self.logger.debug('Connection opened')
        self._add_on_connection_close_callback()
        for name, consumer in self._consumers.items():
            consumer._connection = self._connection
            consumer.open_channel()

    def _add_on_connection_close_callback(self):
        self.logger.debug('Adding connection close callback')
        self._connection.add_on_close_callback(self._on_connection_closed)

    def _on_connection_closed(self, _, reply_code, reply_text):
        if self._closing:
            self._connection.ioloop.stop()

    def _reconnect(self):
        # This is the old connection IOLoop instance, stop its ioloop
        self._connection.ioloop.stop()

        if not self._closing:
            # Create a new connection
            self._connection = self.connect()

            # There is now a new connection, needs a new ioloop to run
            self._connection.ioloop.start()

    def _close_connection(self):
        """This method closes the connection to RabbitMQ.

        """
        self.logger.debug('Closing connection')
        self._connection.close()



import logging
import multiprocessing
import queue
import threading
import time

from gromozeka.concurrency import commands
from gromozeka.concurrency.worker import ThreadWorker


class Pool(threading.Thread):
    """Pool of workers

    Attributes:
        logger(:obj:`logging.Logger`): Class logger
        in_queue(:obj:`multiprocessing.Queue`): Queue to receive input requests
        out_queue(:obj:`multiprocessing.Queue`): Result queue
        cmd_in_queue(:obj:`multiprocessing.Queue`): Command queue
        cmd_out_queue(:obj:`multiprocessing.Queue`): Command result queue
        max_workers(:obj:`int`): Number of workers to start
        worker_class(:class:`gromozeka.concurrency.Worker`): worker class
        workers(:obj:`list` of :obj:`gromozeka.concurrency.Worker`): list of worker instances
        _stop_event(:obj:`multiprocessing.Event`): Stop event

    Args:
        max_workers(:obj:`int`, optional): Number of workers to start
        worker_class(:class:`gromozeka.concurrency.Worker`, optional): Worker class
    """

    __slots__ = ['logger', 'init_max_workers', 'worker_class', 'in_queue', 'out_queue', 'cmd_in_queue', 'cmd_out_queue',
                 '_stop_event', 'workers', 'is_on_stop']

    def __init__(self, max_workers=1, worker_class=ThreadWorker, logger=None):
        self.logger = logger or logging.getLogger("gromozeka.pool")
        self.init_max_workers = max_workers
        self.worker_class = worker_class
        self.in_queue = multiprocessing.Queue(self.init_max_workers)
        self.out_queue = multiprocessing.Queue()
        self.cmd_in_queue = multiprocessing.Queue()
        self.cmd_out_queue = multiprocessing.Queue()
        self._stop_event = multiprocessing.Event()
        self.is_on_stop = False
        self.workers = {}
        super().__init__()

    def listen_cmd(self):
        """Listen to commands

        """
        while not self._stop_event.is_set():
            try:
                command, args = self.cmd_in_queue.get(True, 0.05)
            except queue.Empty:
                continue
            if command == commands.POOL_SHRINK:
                self._shrink(args)
            elif command == commands.POOL_GROW:
                self._grow(args)
            elif command == commands.POOL_REMOVE_WORKER:
                self._remove_worker(args)
            elif command == commands.POOL_SIZE:
                self._size()
            elif command == commands.POOL_STOP:
                self.stop_()

    def start(self):
        """Start pool

        """
        super().start()
        return self.cmd_out_queue.get()

    @property
    def cmd(self):
        """Command queue alias

        Returns:
            multiprocessing.Queue: Command queue
        """
        return self.cmd_in_queue

    def run(self):
        """Start :attr:`workers` in threads or processes

        """
        self.logger.info("start")
        self.grow(self.init_max_workers)
        self.listen_cmd()

    def stop(self):
        """Stop pool and child :attr:`workers`

        """
        self.cmd.put(commands.pool_stop())

    def stop_(self):
        self.is_on_stop = True
        self.out_queue.cancel_join_thread()
        self.in_queue.cancel_join_thread()
        self._shrink(len(self.workers))
        self._stop_event.set()
        self.logger.info("stop")

    def grow(self, n):
        """Grow pool by `n` number of workers

        Args:
            n(int): Number to grow
        """
        self.cmd.put(commands.pool_grow(n))

    def _grow(self, n):
        starting_workers = []
        for _ in range(n):
            worker = self.worker_class(pool=self)
            worker.start()
            self.workers[worker.ident] = worker
            starting_workers.append(worker)
        self._wait_for_start(starting_workers)
        self.in_queue._maxsize = len(self.workers)
        from gromozeka import get_app
        get_app().broker.on_pool_size_change()

    def shrink(self, n):
        """Shrink pool by `n` number of workers

        Args:
            n(int): Number to grow
        """
        self.cmd.put(commands.pool_shrink(n))

    def _shrink(self, n):
        stopping_workers = []
        for _ in range(n):
            ident = list(self.workers.keys())[-1]
            worker = self.workers.pop(ident)
            worker.stop()
            stopping_workers.append(worker)
        self._wait_for_stop(stopping_workers)
        self.in_queue._maxsize = len(self.workers)
        if self.is_on_stop:
            return
        from gromozeka import get_app
        get_app().broker.on_pool_size_change()

    def size(self):
        """Get pool size

        Returns:
            int: Number of workers in pool
        """
        self.cmd.put(commands.pool_size())
        return self.cmd_out_queue.get()

    def _size(self):
        self.cmd_out_queue.put(len(self.workers))

    def remove_worker(self, worker_ident):
        """Remove worker from pool by it`s ident

        Args:
            worker_ident: Worker identification
        """
        self.cmd.put(commands.pool_remove_worker(worker_ident))

    def _remove_worker(self, worker_ident):
        del self.workers[worker_ident]
        self.in_queue._maxsize = len(self.workers)
        from gromozeka import get_app
        get_app().broker.on_pool_size_change()

    def _wait_for_stop(self, workers):
        """Wait until workers stop

        Args:
            workers(:obj:`list` of :obj:`concurrency.Worker`):
        """
        while workers:
            for worker in workers:
                if not worker.is_alive():
                    workers.remove(worker)
                time.sleep(0.05)
        self.cmd_out_queue.put(True)

    def _wait_for_start(self, workers):
        """Wait until workers start

        Args:
            workers(:obj:`list` of :obj:`concurrency.Worker`):
        """
        while workers:
            for worker in workers:
                if worker.is_alive():
                    workers.remove(worker)
                time.sleep(0.05)
        self.cmd_out_queue.put(True)



import functools
import logging
import multiprocessing
import queue
import threading

from gromozeka.exceptions import Retry, MaxRetriesExceedException
from gromozeka.primitives import Task


class Worker:
    """Worker class for `gromozeka.concurrency.Pool`. Uses as mixin.

    Attributes:
        logger(:obj:`logging.Logger`): class logger
        pool(:obj:`gromozeka.concurrency.Pool`): parent pool instance
        _stop_event(:obj:`multiprocessing.Event` or :obj:`threading.Event`)

    Args:
        pool(gromozeka.concurrency.Pool): Parent pool instance
    """

    name = None  # multiprocessing or threading part
    ident = None  # multiprocessing or threading part

    def __init__(self, pool):
        self.logger = logging.getLogger("gromozeka.pool.worker")
        self.pool = pool
        self._stop_event = None

    def run(self):
        """Run worker in thread or process

        Raises:
            exceptions.MaxRetriesExceedException:
            BaseException:
        """
        self.logger.info("start")
        while not self._stop_event.is_set():
            try:
                # get request object from queue
                request = self.pool.in_queue.get(True, 0.05)
            except queue.Empty:
                continue
            # make Task object from request
            task = Task.from_request(request)
            task.on_receive()
            try:
                functools.partial(task.func, *(task, *task.args) if task.bind else task.args, **task.kwargs)()
                # TODO Need result backed to use result queue
                # self.pool.out_queue.put(res)
                task.on_success()
            except Retry as e:
                if task.on_retry(e):
                    continue
                self.pool.remove_worker(self.ident)
                task.on_fail(MaxRetriesExceedException())
            except Exception as e:
                self.pool.remove_worker(self.ident)
                task.on_fail(e)
        self.logger.info("stop")

    def stop(self):
        """Stop worker. multiprocessing or threading part

        """
        pass

    def is_alive(self):
        """Check worker is alive. multiprocessing or threading part

        Returns:
            bool: True if worker is alive or False
        """
        pass


class ProcessWorker(multiprocessing.Process, Worker):
    """See `gromozeka.concurrency.Worker` documentation

    Attributes:
        _stop_event(:obj:`multiprocessing.Event`): Stop event
    """
    __doc__ = Worker.__doc__

    def __init__(self, pool):
        multiprocessing.Process.__init__(self)
        Worker.__init__(self, pool=pool)
        self._stop_event = multiprocessing.Event()

    def run(self):
        Worker.run(self)

    def stop(self, timeout=None):
        self._stop_event.set()
        super().join(timeout)


class ThreadWorker(threading.Thread, Worker):
    """See `gromozeka.concurrency.Worker` documentation

    Attributes:
        _stop_event(:obj:`threading.Event`): Stop event
    """
    __doc__ = Worker.__doc__

    def __init__(self, pool):
        threading.Thread.__init__(self)
        Worker.__init__(self, pool=pool)
        self._stop_event = threading.Event()

    def run(self):
        Worker.run(self)

    def stop(self, timeout=None):
        self._stop_event.set()
        super().join(timeout)



from gromozeka.primitives.protocol import TaskType, SchedulerType, BrokerPointType, MessageType, Request
from gromozeka.primitives.base import Task, RegistryTask, task

__all__ = ['Task', 'RegistryTask', 'task', 'TaskType', 'SchedulerType', 'BrokerPointType', 'MessageType', 'Request']



import logging
import os
from logging import config

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'console': {
            'format': '%(asctime)-15s %(levelname)s %(name)s [%(threadName)s %(processName)s] %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'console'
        },
    },
    'loggers': {
        'gromozeka': {'handlers': ['console'], 'level': 'INFO'},
        'gromozeka.pool': {},
        'gromozeka.pool.worker': {},
        'gromozeka.scheduler': {},
        'gromozeka.broker': {},
        'gromozeka.broker.consumer': {},
    }
}


class Config:
    """Configuration class
    
    Attributes:
        debug(:obj:`bool`): Set logger level. Default: `False`
        broker_reconnect_max_retries(:obj:`int`): Broker reconnect attempts. Default: 10
        broker_reconnect_retry_countdown(:obj:`int`): Sleep time between reconnects (seconds). Default: 10
        broker_retry_exchange(:obj:`str`): Broker exchange for retries. Default: 'countdown_exchange'
        broker_retry_queue(:obj:`str`): Broker queue name for retries. Default: 'countdown_queue'
        broker_retry_routing_key(:obj:`str`): Broker routing key for retries. Default: 'countdown'
        broker_eta_exchange(:obj:`str`): Broker exchange for eta tasks. Default: 'eta_exchange'
        broker_eta_queue(:obj:`str`): Broker queue for retries. Default: 'eta_queue'
        broker_eta_routing_key(:obj:`str`): Broker routing key for retries. Default: 'eta'
        broker_url(:obj:`str`): Broker url. Default: 'amqp://guest:guest@localhost:5672/%2F'
    """

    def __init__(self):
        self.debug = False
        self.broker_reconnect_max_retries = 10
        self.broker_reconnect_retry_countdown = 10
        self.broker_retry_exchange = 'countdown_exchange'
        self.broker_retry_queue = 'countdown_queue'
        self.broker_retry_routing_key = 'countdown'
        self.broker_eta_exchange = 'eta_exchange'
        self.broker_eta_queue = 'eta_queue'
        self.broker_eta_routing_key = 'eta'
        self.broker_url = 'amqp://guest:guest@localhost:5672/%2F'
        config.dictConfig(LOGGING)

    def from_env(self):
        """Configure Gromozeka with environment variables

        Returns:
            gromozeka.Config: Configured object
        """

        for name in dir(self):
            if name.startswith('__') or callable(getattr(self, name)):
                continue
            new_attr = os.getenv(name, getattr(self, name))
            type_ = type(getattr(self, name))
            self._set_var_type(type_, name, new_attr)
        if self.debug:
            logging.getLogger('gromozeka').setLevel(logging.DEBUG)

        return self

    def from_dict(self, config_dict):
        """Configure Gromozeka with :obj:`dict`

        Args:
            config_dict(dict): config :obj:`dict`

        Returns:
            gromozeka.Config: Configured object
        """
        for name, value in config_dict.items():
            if not hasattr(self, name):
                from gromozeka import GromozekaException
                raise GromozekaException('Unknown config option `{}`'.format(name))
            type_ = type(getattr(self, name))
            self._set_var_type(type_, name, value)
        if self.debug:
            logging.getLogger('gromozeka').setLevel(logging.DEBUG)

        return self

    def _set_var_type(self, type_, name, attr):
        if type(None) == type_:
            setattr(self, name, attr)
        if type(False) == type_:
            if attr in ('0', 0, 'false', 'False', ''):
                setattr(self, name, False)
            else:
                setattr(self, name, True)
        else:
            setattr(self, name, type_(attr))


app_config = Config()

__all__ = [app_config]



"""Commands to synchronize processes or threads

"""

from typing import NamedTuple, Any


class Command(NamedTuple):
    """Helper class to make commands

    """
    # command number. Find it in `async_task_processor.primitives.commands`
    command: int
    # command arguments
    args: Any


POOL_START = 1
POOL_STOP = 2
POOL_GROW = 3
POOL_SHRINK = 4
POOL_SIZE = 5
POOL_REMOVE_WORKER = 6
BROKER_ADD_CONSUMER = 7
BROKER_PUBLISH = 8
BROKER_ACK = 9
BROKER_REJECT = 10
BROKER_PREFETCH_COUNT = 11


# POOL
def pool_stop():
    """Command to stop pool

    Returns:
        Command:
    """
    return Command(command=POOL_STOP, args=None)


def pool_grow(n):
    """Command to grow pool

    Args:
        n: Number of workers to grow

    Returns:
        Command:
    """
    return Command(command=POOL_GROW, args=n)


def pool_shrink(n):
    """Command to shrink pool

    Args:
        n: Number of workers to shrink

    Returns:
        Command:
    """
    return Command(command=POOL_SHRINK, args=n)


def pool_size():
    """Command to get pool size

    Returns:
        Command:
    """
    return Command(command=POOL_SIZE, args=None)


def pool_remove_worker(worker_ident):
    """Command to remove worker from pool

    Args:
        worker_ident(int): Worker identification

    Returns:
        Command:
    """
    return Command(command=POOL_REMOVE_WORKER, args=worker_ident)


# BROKER
def broker_add_consumer(task_id, broker_point):
    """Command to add new consumer with `broker_point` to task

    Args:
        task_id(str): Unique task identification
        broker_point(primitives.BrokerPointType): Broker entry

    Returns:
        Command:
    """
    return Command(command=BROKER_ADD_CONSUMER, args=dict(task_id=task_id, broker_point=broker_point))


def broker_publish(request):
    """Command to publish `primitives.Request` to customer

    Args:
        request(primitives.Request): Request to publish

    Returns:
        Command:
    """
    return Command(command=BROKER_PUBLISH, args=request)


def broker_ack(broker_point, delivery_tag):
    """Command to acknowledge task message by it's broker point, delivery_tag or scheduler_tag

    Args:
        broker_point(primitives.BrokerPointType): Broker entry
        delivery_tag(int): Task delivery tag
    Returns:
        Command:
    """
    return Command(command=BROKER_ACK, args={'broker_point': broker_point, 'delivery_tag': delivery_tag})


def broker_reject(broker_point, delivery_tag):
    """Command to reject task message by it's broker point, delivery_tag or scheduler_tag

    Args:
        broker_point(primitives.BrokerPointType): Broker entry
        delivery_tag(int): Task delivery tag

    Returns:
        Command:
    """
    return Command(command=BROKER_REJECT, args={'broker_point': broker_point, 'delivery_tag': delivery_tag})


def broker_prefetch_count(broker_point, new_size):
    """Command to change prefetch_count

    Args:
        broker_point(primitives.BrokerPointType): Broker entry
        new_size(int): New prefetch_count size

    Returns:
        Command:
    """
    return Command(command=BROKER_PREFETCH_COUNT, args={'broker_point': broker_point, 'new_size': new_size})



import io
import re
from os.path import join, dirname

from pip.req import parse_requirements
from setuptools import setup, find_packages

PACKAGE = "gromozeka"
NAME = "gromozeka"
DESCRIPTION = "Distributed task queue."
AUTHOR = "Klimov Konstantin"
AUTHOR_EMAIL = "moelius1983@gmail.com"
URL = "https://github.com/moelius/gromozeka"

with io.open('gromozeka/__init__.py', 'rt', encoding='utf8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setup(
    name=NAME,
    version=version,
    description=DESCRIPTION,
    long_description=open(join(dirname(__file__), 'README.rst')).read(),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    keywords='task job queue distributed messaging pika rabbit rabbitmq pool worker',
    license="MIT",
    url=URL,
    packages=find_packages(exclude=["tests.*", "tests"]),
    install_requires=[str(ir.req) for ir in parse_requirements('requirements.txt', session='session')],
    extras_require={
        'docs': [str(ir.req) for ir in parse_requirements('requirements/docs.txt', session='session')],
    },
    zip_safe=False,
)



import json
from typing import NamedTuple

import gromozeka.brokers.rabbit


class TaskType(NamedTuple):
    """Helper class

    """
    uuid: str
    id: str
    args: list = []
    kwargs: dict = {}
    state: str = 'PENDING'
    retries: int = 0
    delivery_tag: int = 0


class BrokerPointType(NamedTuple):
    """Helper class

    """
    exchange: str
    exchange_type: str
    queue: str
    routing_key: str


# TODO eta,every,interval for periodic tasks - not ready yet (
class SchedulerType(NamedTuple):
    """Helper class

    """
    eta: str = ''
    every: str = ''
    interval: str = ''
    at: str = ''
    countdown: str = ''


class MessageType(NamedTuple):
    """Helper class

    """
    task: TaskType
    scheduler: SchedulerType = None
    broker_point: BrokerPointType = None


class Request:
    """Request class. Use as task representation for Broker

    Attributes:
        task(gromozeka.primitives.TaskType): Task tuple
        broker_point(gromozeka.primitives.BrokerPointType): Broker entry tuple
        scheduler(gromozeka.primitives.SchedulerType): Scheduler tuple

    Args:
        task(gromozeka.primitives.TaskType): Task tuple
        broker_point(gromozeka.primitives.BrokerPointType): Broker entry tuple
        scheduler(gromozeka.primitives.SchedulerType): Scheduler tuple
    """
    __slots__ = ['task', 'broker_point', 'scheduler']

    def __init__(self, task, broker_point, scheduler=None):
        self.task = task
        self.broker_point = broker_point
        self.scheduler = scheduler

    def to_json(self):
        """Convert `Request` to json

        Returns:
            Json representation of `Request`
        """
        return json.dumps(MessageType(task=self.task, broker_point=self.broker_point, scheduler=self.scheduler))

    @classmethod
    def from_json(cls, json_message, state=None, delivery_tag=None):
        """Make `Request` from json

        Args:
            json_message(:obj:`str`, optional): `Request` json string
            state(:obj:`str`, optional): `TaskType` new state
            delivery_tag(:obj:`int`, optional): Original mess

        Returns:
            primitives.Request: `Request` object
        """
        request = json.loads(json_message)

        task_raw = request[0]
        task_uuid = task_raw[0]
        task_id = task_raw[1]
        task_args = task_raw[2]
        task_kwargs = task_raw[3]
        task_state = task_raw[4] if len(task_raw) > 4 else None
        task_retries = task_raw[5] if len(task_raw) > 5 else None
        task_delivery_tag = task_raw[6] if len(task_raw) > 6 else None
        task = TaskType(
            uuid=task_uuid,
            id=task_id,
            state=state or task_state,
            args=task_args,
            kwargs=task_kwargs,
            retries=task_retries,
            delivery_tag=delivery_tag or task_delivery_tag)

        if len(request) > 1 and request[1]:
            scheduler = SchedulerType(*request[1])
        else:
            scheduler = None

        if len(request) > 2 and request[2]:
            broker_point = BrokerPointType(*request[2])
        else:
            broker_point = gromozeka.get_app().get_task(task_id).broker_point
        return cls(task=task, broker_point=broker_point, scheduler=scheduler)
